from textwrap import dedent
from unittest import TestCase

from pcs.common.reports import codes as report_codes
from pcs.lib.file import toolbox
from pcs.lib.interface.config import ParserErrorException
from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal


class JsonParser(TestCase):
    file_type_code = "file type code"
    file_path = "file path"

    def test_success(self):
        self.assertEqual(
            toolbox.JsonParser.parse(
                dedent(
                    """
                {
                    "simple": "value",
                    "list": ["item1", "item2"]
                }
                """
                )
            ),
            {
                "simple": "value",
                "list": ["item1", "item2"],
            },
        )

    def _parse_error(self, force_code, is_forced):
        with self.assertRaises(ParserErrorException) as cm:
            toolbox.JsonParser.parse(
                dedent(
                    """
                {
                    "simple": "value",
                    "list": ["item1" "item2"]
                }
                """
                )
            )
        return toolbox.JsonParser.exception_to_report_list(
            cm.exception,
            self.file_type_code,
            self.file_path,
            force_code,
            is_forced,
        )

    def _parse_error_report_args(self):
        return dict(
            file_path=self.file_path,
            file_type_code=self.file_type_code,
            line_number=4,
            column_number=22,
            position=47,
            reason="Expecting ',' delimiter",
            full_msg="Expecting ',' delimiter: line 4 column 22 (char 47)",
        )

    def test_parse_error(self):
        assert_report_item_list_equal(
            self._parse_error(None, False),
            [
                fixture.error(
                    report_codes.PARSE_ERROR_JSON_FILE,
                    **self._parse_error_report_args(),
                ),
            ],
        )

    def test_parse_error_forcible(self):
        assert_report_item_list_equal(
            self._parse_error("force code", False),
            [
                fixture.error(
                    report_codes.PARSE_ERROR_JSON_FILE,
                    force_code="force code",
                    **self._parse_error_report_args(),
                ),
            ],
        )

    def test_parse_error_forced(self):
        assert_report_item_list_equal(
            self._parse_error("force code", True),
            [
                fixture.warn(
                    report_codes.PARSE_ERROR_JSON_FILE,
                    **self._parse_error_report_args(),
                ),
            ],
        )


class JsonExporter(TestCase):
    def test_success(self):
        self.assertEqual(
            toolbox.JsonExporter.export(
                {
                    "simple": "value",
                    "list": ["item1", "item2"],
                }
            ),
            dedent(
                """\
                {
                    "list": [
                        "item1",
                        "item2"
                    ],
                    "simple": "value"
                }
            """.rstrip()
            ).encode("utf-8"),
        )



from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal

from pcs.common.reports import codes as report_codes
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.booth import config_parser
from pcs.lib.booth.config_parser import ConfigItem


class BuildTest(TestCase):
    def test_build_file_content_from_parsed_structure(self):
        self.assertEqual(
            "\n".join(
                [
                    "authfile = /path/to/auth.file",
                    "site = 1.1.1.1",
                    "site = 2.2.2.2",
                    "arbitrator = 3.3.3.3",
                    'ticket = "TA"',
                    'ticket = "TB"',
                    "  timeout = 10",
                    "",  # newline at the end
                ]
            ).encode("utf-8"),
            config_parser.Exporter.export(
                [
                    ConfigItem("authfile", "/path/to/auth.file"),
                    ConfigItem("site", "1.1.1.1"),
                    ConfigItem("site", "2.2.2.2"),
                    ConfigItem("arbitrator", "3.3.3.3"),
                    ConfigItem("ticket", "TA"),
                    ConfigItem("ticket", "TB", [ConfigItem("timeout", "10")]),
                ]
            ),
        )


class OrganizeLinesTest(TestCase):
    def test_move_non_ticket_config_keys_above_tickets(self):
        self.assertEqual(
            [
                ConfigItem("site", "1.1.1.1"),
                ConfigItem("site", "2.2.2.2"),
                ConfigItem("arbitrator", "3.3.3.3"),
                ConfigItem("ticket", "TA"),
            ],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._organize_lines(
                [
                    ("site", "1.1.1.1"),
                    ("ticket", "TA"),
                    ("site", "2.2.2.2"),
                    ("arbitrator", "3.3.3.3"),
                ]
            ),
        )

    def test_use_ticket_key_as_ticket_detail(self):
        # pylint: disable=invalid-name
        self.maxDiff = None
        self.assertEqual(
            [
                ConfigItem("site", "1.1.1.1"),
                ConfigItem("expire", "300"),
                ConfigItem("site", "2.2.2.2"),
                ConfigItem("arbitrator", "3.3.3.3"),
                ConfigItem(
                    "ticket",
                    "TA",
                    [
                        ConfigItem("timeout", "10"),
                        ConfigItem("--nonexistent", "value"),
                        ConfigItem("expire", "300"),
                    ],
                ),
                ConfigItem(
                    "ticket",
                    "TB",
                    [
                        ConfigItem("timeout", "20"),
                        ConfigItem("renewal-freq", "40"),
                    ],
                ),
            ],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._organize_lines(
                [
                    ("site", "1.1.1.1"),
                    ("expire", "300"),  # out of ticket content is kept global
                    ("ticket", "TA"),
                    ("site", "2.2.2.2"),  # move to global
                    ("timeout", "10"),
                    (
                        "--nonexistent",
                        "value",
                    ),  # no global is kept under ticket
                    ("expire", "300"),
                    ("ticket", "TB"),
                    ("arbitrator", "3.3.3.3"),
                    ("timeout", "20"),
                    ("renewal-freq", "40"),
                ]
            ),
        )


class ParseRawLinesTest(TestCase):
    def test_parse_simple_correct_lines(self):
        self.assertEqual(
            [
                ("site", "1.1.1.1"),
                ("site", "2.2.2.2"),
                ("arbitrator", "3.3.3.3"),
                ("syntactically_correct", "nonsense"),
                ("line-with", "hash#literal"),
            ],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines(
                "\n".join(
                    [
                        "site = 1.1.1.1",
                        " site  =  2.2.2.2 ",
                        "arbitrator=3.3.3.3",
                        "syntactically_correct = nonsense",
                        "line-with = hash#literal",
                        "",
                    ]
                )
            ),
        )

    def test_parse_lines_with_whole_line_comment(self):
        self.assertEqual(
            [("site", "1.1.1.1")],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines(
                "\n".join(
                    [
                        " # some comment",
                        "site = 1.1.1.1",
                    ]
                )
            ),
        )

    def test_skip_empty_lines(self):
        self.assertEqual(
            [("site", "1.1.1.1")],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines(
                "\n".join(
                    [
                        " ",
                        "site = 1.1.1.1",
                    ]
                )
            ),
        )

    def test_raises_when_unexpected_lines_appear(self):
        invalid_line_list = [
            "first invalid line",
            "second = 'invalid line' something else #comment",
            "third = 'invalid line 'something#'#",
        ]
        line_list = ["site = 1.1.1.1"] + invalid_line_list
        with self.assertRaises(config_parser.InvalidLines) as context_manager:
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines("\n".join(line_list))
        self.assertEqual(context_manager.exception.args[0], invalid_line_list)

    def test_parse_lines_finishing_with_comment(self):
        self.assertEqual(
            [("site", "1.1.1.1")],
            # testing a function which should not be used outside of the module
            # pylint: disable=protected-access
            config_parser._parse_to_raw_lines(
                "\n".join(
                    [
                        "site = '1.1.1.1' #comment",
                    ]
                )
            ),
        )


class ParseTest(TestCase):
    def test_raises_when_invalid_lines_appear(self):
        invalid_line_list = [
            "first invalid line",
            "second = 'invalid line' something else #comment",
        ]
        line_list = ["site = 1.1.1.1"] + invalid_line_list

        with self.assertRaises(ParserErrorException) as cm:
            config_parser.Parser.parse("\n".join(line_list).encode("utf-8"))

        assert_report_item_list_equal(
            config_parser.Parser.exception_to_report_list(
                cm.exception, "does not matter", "path", None, False
            ),
            [
                fixture.error(
                    report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    line_list=invalid_line_list,
                    file_path="path",
                ),
            ],
        )
        assert_report_item_list_equal(
            config_parser.Parser.exception_to_report_list(
                cm.exception,
                "does not matter",
                "path",
                report_codes.FORCE,
                False,
            ),
            [
                fixture.error(
                    report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    force_code=report_codes.FORCE,
                    line_list=invalid_line_list,
                    file_path="path",
                ),
            ],
        )
        assert_report_item_list_equal(
            config_parser.Parser.exception_to_report_list(
                cm.exception,
                "does not matter",
                "path",
                report_codes.FORCE,
                True,
            ),
            [
                fixture.warn(
                    report_codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    line_list=invalid_line_list,
                    file_path="path",
                ),
            ],
        )

    def test_do_not_raises_when_no_invalid_liens_there(self):
        parsed = config_parser.Parser.parse("site = 1.1.1.1".encode("utf-8"))
        self.assertEqual(len(parsed), 1)
        self.assertEqual(parsed[0].key, "site")
        self.assertEqual(parsed[0].value, "1.1.1.1")
        self.assertEqual(parsed[0].details, [])



from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.misc import get_test_resource as rc

import pcs.lib.corosync.config_facade as lib
from pcs.lib.corosync.config_parser import Parser


def _read_file(name):
    with open(rc(name)) as a_file:
        return a_file.read()


def _get_facade(config_text):
    return lib.ConfigFacade(Parser.parse(config_text.encode("utf-8")))


class GetQuorumOptionsTest(TestCase):
    def test_no_quorum(self):
        config = ""
        facade = _get_facade(config)
        options = facade.get_quorum_options()
        self.assertEqual({}, options)
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_empty_quorum(self):
        config = dedent(
            """\
            quorum {
            }
        """
        )
        facade = _get_facade(config)
        options = facade.get_quorum_options()
        self.assertEqual({}, options)
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_no_options(self):
        config = dedent(
            """\
            quorum {
                provider: corosync_votequorum
            }
        """
        )
        facade = _get_facade(config)
        options = facade.get_quorum_options()
        self.assertEqual({}, options)
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_some_options(self):
        config = dedent(
            """\
            quorum {
                provider: corosync_votequorum
                wait_for_all: 0
                nonsense: ignored
                auto_tie_breaker: 1
                last_man_standing: 0
                last_man_standing_window: 1000
            }
        """
        )
        facade = _get_facade(config)
        options = facade.get_quorum_options()
        self.assertEqual(
            {
                "auto_tie_breaker": "1",
                "last_man_standing": "0",
                "last_man_standing_window": "1000",
                "wait_for_all": "0",
            },
            options,
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_option_repeated(self):
        config = dedent(
            """\
            quorum {
                wait_for_all: 0
                wait_for_all: 1
            }
        """
        )
        facade = _get_facade(config)
        options = facade.get_quorum_options()
        self.assertEqual(
            {
                "wait_for_all": "1",
            },
            options,
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)

    def test_quorum_repeated(self):
        config = dedent(
            """\
            quorum {
                wait_for_all: 0
                last_man_standing: 0
            }
            quorum {
                last_man_standing_window: 1000
                wait_for_all: 1
            }
        """
        )
        facade = _get_facade(config)
        options = facade.get_quorum_options()
        self.assertEqual(
            {
                "last_man_standing": "0",
                "last_man_standing_window": "1000",
                "wait_for_all": "1",
            },
            options,
        )
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)


class IsEnabledAutoTieBreaker(TestCase):
    def test_enabled(self):
        config = dedent(
            """\
            quorum {
                auto_tie_breaker: 1
            }
        """
        )
        facade = _get_facade(config)
        self.assertTrue(facade.is_enabled_auto_tie_breaker())

    def test_disabled(self):
        config = dedent(
            """\
            quorum {
                auto_tie_breaker: 0
            }
        """
        )
        facade = _get_facade(config)
        self.assertFalse(facade.is_enabled_auto_tie_breaker())

    def test_no_value(self):
        config = dedent(
            """\
            quorum {
            }
        """
        )
        facade = _get_facade(config)
        self.assertFalse(facade.is_enabled_auto_tie_breaker())


class SetQuorumOptionsTest(TestCase):
    @staticmethod
    def get_two_node(facade):
        two_node = None
        for quorum in facade.config.get_sections("quorum"):
            for dummy_name, value in quorum.get_attributes("two_node"):
                two_node = value
        return two_node

    def test_add_missing_section(self):
        config = ""
        facade = _get_facade(config)
        facade.set_quorum_options({"wait_for_all": "0"})
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            dedent(
                """\
                quorum {
                    wait_for_all: 0
                }
            """
            ),
            facade.config.export(),
        )

    def test_del_missing_section(self):
        config = ""
        facade = _get_facade(config)
        facade.set_quorum_options({"wait_for_all": ""})
        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual("", facade.config.export())

    def test_add_all_options(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        expected_options = {
            "auto_tie_breaker": "1",
            "last_man_standing": "0",
            "last_man_standing_window": "1000",
            "wait_for_all": "0",
        }
        facade.set_quorum_options(expected_options)

        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        test_facade = _get_facade(facade.config.export())
        self.assertEqual(expected_options, test_facade.get_quorum_options())

    def test_complex(self):
        config = dedent(
            """\
            quorum {
                wait_for_all: 0
                last_man_standing_window: 1000
            }
            quorum {
                wait_for_all: 0
                last_man_standing: 1
            }
        """
        )
        facade = _get_facade(config)
        facade.set_quorum_options(
            {
                "auto_tie_breaker": "1",
                "wait_for_all": "1",
                "last_man_standing_window": "",
            }
        )

        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        test_facade = _get_facade(facade.config.export())
        self.assertEqual(
            {
                "auto_tie_breaker": "1",
                "last_man_standing": "1",
                "wait_for_all": "1",
            },
            test_facade.get_quorum_options(),
        )

    def test_2nodes_atb_on(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        self.assertEqual(2, len(facade.get_nodes()))

        facade.set_quorum_options({"auto_tie_breaker": "1"})

        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            "1", facade.get_quorum_options().get("auto_tie_breaker", None)
        )

        two_node = self.get_two_node(facade)
        self.assertTrue(two_node is None or two_node == "0")

    def test_2nodes_atb_off(self):
        config = _read_file("corosync.conf")
        facade = _get_facade(config)
        self.assertEqual(2, len(facade.get_nodes()))

        facade.set_quorum_options({"auto_tie_breaker": "0"})

        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            "0", facade.get_quorum_options().get("auto_tie_breaker", None)
        )

        two_node = self.get_two_node(facade)
        self.assertTrue(two_node == "1")

    def test_3nodes_atb_on(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        self.assertEqual(3, len(facade.get_nodes()))

        facade.set_quorum_options({"auto_tie_breaker": "1"})

        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            "1", facade.get_quorum_options().get("auto_tie_breaker", None)
        )

        two_node = self.get_two_node(facade)
        self.assertTrue(two_node is None or two_node == "0")

    def test_3nodes_atb_off(self):
        config = _read_file("corosync-3nodes.conf")
        facade = _get_facade(config)
        self.assertEqual(3, len(facade.get_nodes()))

        facade.set_quorum_options({"auto_tie_breaker": "0"})

        self.assertTrue(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)
        self.assertEqual(
            "0", facade.get_quorum_options().get("auto_tie_breaker", None)
        )

        two_node = self.get_two_node(facade)
        self.assertTrue(two_node is None or two_node == "0")
